from django.db import models


class PaddleCurrencyCodeField(models.CharField):
    """
    A field used to store a three-letter currency code (eg. USD, EUR, ...)
    """

    def __init__(self, *args, **kwargs):
        defaults = {"max_length": 3, "help_text": "Three-letter ISO currency code"}
        defaults.update(kwargs)
        super().__init__(*args, **defaults)



from distutils.util import strtobool

from django.http import HttpResponse, HttpResponseBadRequest, JsonResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from django.views.generic.edit import BaseCreateView

from . import signals
from .models import Checkout, convert_datetime_strings_to_datetimes
from .utils import is_valid_webhook


@method_decorator(csrf_exempt, name="dispatch")
class PaddleWebhookView(View):
    SUPPORTED_WEBHOOKS = (
        # Subscription Alerts
        "subscription_created",
        "subscription_updated",
        "subscription_cancelled",
        "subscription_payment_succeeded",
        "subscription_payment_failed",
        "subscription_payment_refunded",
        # One-off Purchases
        "locker_processed",
        "payment_succeeded",
        "payment_refunded",
        # Risk & Dispute Alerts
        "payment_dispute_created",
        "payment_dispute_closed",
        "high_risk_transaction_created",
        "high_risk_transaction_updated",
        # Payout Alerts
        "transfer_created",
        "transfer_paid",
        # Audience Alerts
        "new_audience_member",
        "update_audience_member",
    )

    def post(self, request, *args, **kwargs):
        """
        handle paddle webhook requests by
        - validating the payload signature
        - sending a django signal for each of the SUPPORTED_WEBHOOKS
        """
        payload = request.POST.dict()

        if not is_valid_webhook(payload):
            return HttpResponseBadRequest("webhook validation failed")

        alert_name = payload.get("alert_name")
        if not alert_name:
            return HttpResponseBadRequest("'alert_name' missing")

        if alert_name in self.SUPPORTED_WEBHOOKS:
            signal = getattr(signals, alert_name)
            if signal:  # pragma: no cover
                signal.send(sender=self.__class__, payload=payload)

        return HttpResponse()


class PaddlePostCheckoutApiView(BaseCreateView):
    http_method_names = ["post"]

    def post(self, request, *args, **kwargs):
        data = request.POST.dict()
        redirect_url = data.pop("redirect_url") if "redirect_url" in data else ""
        pk = data.pop("id")
        if not pk:
            return HttpResponseBadRequest('Missing "id"')
        try:
            data["completed"] = bool(strtobool(data["completed"]))
        except (KeyError, ValueError):
            return HttpResponseBadRequest('Missing "completed"')

        try:
            data = convert_datetime_strings_to_datetimes(data, Checkout)
        except ValueError as e:
            return HttpResponseBadRequest(e)

        Checkout.objects.update_or_create(pk=pk, defaults=data)

        next_url = request.GET.get("next")
        if next_url:
            next_url = "{0}?checkout={1}".format(next_url, pk)
            return JsonResponse({"redirect_url": next_url}, status=200)

        if redirect_url:
            redirect_url = "{0}?checkout={1}".format(redirect_url, pk)
            return JsonResponse({"redirect_url": redirect_url}, status=200)

        return JsonResponse({}, status=204)


paddle_webhook_view = PaddleWebhookView.as_view()
post_checkout_api_view = PaddlePostCheckoutApiView.as_view()



# Generated by Django 3.0.2 on 2020-09-06 14:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djpaddle', '0002_checkout'),
    ]

    operations = [
        migrations.AlterField(
            model_name='checkout',
            name='id',
            field=models.CharField(max_length=64, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='checkout_id',
            field=models.CharField(max_length=64),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='id',
            field=models.CharField(max_length=64, primary_key=True, serialize=False),
        ),
    ]



import logging
from datetime import datetime

from django.db import models
from django.conf import settings as djsettings
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from paddle import PaddleClient

from . import settings, signals, mappers
from .fields import PaddleCurrencyCodeField
from .utils import PADDLE_DATETIME_FORMAT, PADDLE_DATE_FORMAT

log = logging.getLogger("djpaddle")

paddle_client = PaddleClient(vendor_id=settings.DJPADDLE_VENDOR_ID, api_key=settings.DJPADDLE_API_KEY)


class PaddleBaseModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Plan(PaddleBaseModel):
    """
    'Plan' represents a Paddle subscription plan.
    """

    PADDLE_URI_LIST = "subscription/plans"

    BILLING_TYPE_DAY = "day"
    BILLING_TYPE_MONTH = "month"
    BILLING_TYPE_YEAR = "year"
    BILLING_TYPE = (
        (BILLING_TYPE_DAY, _("day")),
        (BILLING_TYPE_MONTH, _("month")),
        (BILLING_TYPE_YEAR, _("year")),
    )

    name = models.CharField(max_length=255)
    billing_type = models.CharField(choices=BILLING_TYPE, max_length=255)
    billing_period = models.IntegerField()
    trial_days = models.IntegerField(default=0)

    class Meta:
        ordering = ["created_at"]

    @classmethod
    def api_list(cls):
        return paddle_client.list_plans()

    @classmethod
    def api_get(cls, plan_id):
        plan_data = paddle_client.list_plans(plan=int(plan_id))[0]
        return plan_data

    @classmethod
    def sync_from_paddle_data(cls, data):
        pk = data.pop("id")
        initial_price = data.pop("initial_price", {})
        recurring_price = data.pop("recurring_price", {})

        plan, __ = cls.objects.get_or_create(pk=pk, defaults=data)

        # drop all existing prices and recreate them
        plan.prices.all().delete()
        prices = []
        for currency, quantity in initial_price.items():
            price = Price(
                plan=plan,
                currency=currency,
                quantity=float(quantity),
                recurring=False,
            )
            prices.append(price)
        for currency, quantity in recurring_price.items():
            price = Price(
                plan=plan,
                currency=currency,
                quantity=float(quantity),
                recurring=True,
            )
            prices.append(price)

        Price.objects.bulk_create(prices)

        return plan

    def __str__(self):
        return "{}:{}".format(self.name, self.id)


class Price(PaddleBaseModel):
    plan = models.ForeignKey("djpaddle.Plan", on_delete=models.CASCADE, related_name="prices")
    currency = PaddleCurrencyCodeField()
    quantity = models.FloatField()
    recurring = models.BooleanField()

    def __str__(self):
        return "{} {}".format(self.quantity, self.currency)

    class Meta:
        ordering = ["currency", "recurring"]
        unique_together = ("plan", "currency", "recurring")


class Subscription(PaddleBaseModel):
    """
    'Subscription' represents a Paddle subscription.

    They are automatically updated on behalf of the webhook calls
    'subscription_created', 'subscription_updated' and 'subscription_cancelled'.

    Stale subscriptions that are not associated with a subscriber are being
    linked by a customizable getter function. Linking is optional and can be configured through the
    settings ('DJPADDLE_LINK_STALE_SUBSCRIPTIONS').
    """

    STATUS_ACTIVE = "active"
    STATUS_TRIALING = "trialing"
    STATUS_PAST_DUE = "past_due"
    STATUS_PAUSED = "paused"
    STATUS_DELETED = "deleted"
    STATUS_CHOICES = (
        (STATUS_ACTIVE, _("active")),
        (STATUS_TRIALING, _("trialing")),
        (STATUS_PAST_DUE, _("past due")),
        (STATUS_PAUSED, _("paused")),
        (STATUS_DELETED, _("deleted")),
    )

    id = models.CharField(max_length=64, primary_key=True)
    subscriber = models.ForeignKey(
        settings.DJPADDLE_SUBSCRIBER_MODEL,
        related_name="subscriptions",
        null=True,
        default=None,
        on_delete=models.CASCADE,
    )

    cancel_url = models.URLField()
    checkout_id = models.CharField(max_length=64)
    currency = models.CharField(max_length=3)
    email = models.EmailField()
    event_time = models.DateTimeField()
    marketing_consent = models.BooleanField()
    next_bill_date = models.DateTimeField()
    passthrough = models.TextField()
    quantity = models.IntegerField()
    source = models.URLField()
    status = models.CharField(choices=STATUS_CHOICES, max_length=16)
    plan = models.ForeignKey(Plan, on_delete=models.CASCADE)
    unit_price = models.FloatField()
    update_url = models.URLField()

    class Meta:
        ordering = ["created_at"]

    @classmethod
    def _sanitize_webhook_payload(cls, payload):
        data = {}
        data["id"] = payload.pop("subscription_id")

        Subscriber = settings.get_subscriber_model()
        try:
            data["subscriber"] = mappers.get_subscriber_by_payload(Subscriber, payload)
        except Subscriber.DoesNotExist:
            warning = "Subscriber could not be found for subscription {0} with payload {1}. Subscriber left empty."
            log.warning(warning.format(data["id"], payload))
            data["subscriber"] = None

        # transform `subscription_plan_id` to plan ref
        plan_id = payload.pop("subscription_plan_id")
        try:
            data["plan"] = Plan.objects.get(pk=plan_id)
        except Plan.DoesNotExist:
            plan_data = Plan.api_get(plan_id=plan_id)
            data["plan"] = Plan.sync_from_paddle_data(plan_data)

        # sanitize fields
        valid_field_names = [field.name for field in cls._meta.get_fields()]
        for key, value in payload.items():
            # e.g. 'new_status' --> 'status'
            if key.startswith("new_"):
                key = key[4:]

            if key in valid_field_names:
                data[key] = value

        data = convert_datetime_strings_to_datetimes(data, cls)

        return data

    @classmethod
    def create_or_update_by_payload(cls, payload):
        data = cls._sanitize_webhook_payload(payload)
        pk = data.get("id")
        try:
            subscription = cls.objects.get(pk=pk)
        except cls.DoesNotExist:
            return cls.objects.create(pk=pk, **data)

        if subscription.event_time < data["event_time"]:
            cls.objects.filter(pk=pk).update(**data)

    def __str__(self):
        return "{}:{}".format(self.subscriber, self.id)


class Checkout(models.Model):
    """
    Used to store checkout info from PaddleJS. Transient model which acts as
    a backup in case the webhook is not recieved straight away
    """

    id = models.CharField(max_length=64, primary_key=True)
    completed = models.BooleanField(null=True)
    passthrough = models.TextField(null=True, blank=True)
    email = models.EmailField(null=True, blank=True)
    created_at = models.DateTimeField(null=True, blank=True)


@receiver(signals.subscription_created)
@receiver(signals.subscription_updated)
@receiver(signals.subscription_cancelled)
@receiver(signals.subscription_payment_succeeded)
def subscription_event(sender, payload, *args, **kwargs):
    Subscription.create_or_update_by_payload(payload)


if settings.DJPADDLE_LINK_STALE_SUBSCRIPTIONS:

    @receiver(post_save, sender=settings.DJPADDLE_SUBSCRIBER_MODEL)
    def link_stale_subscriptions_to_subscriber(sender, instance, created, *args, **kwargs):
        if created:
            queryset = Subscription.objects.filter(subscriber=None)
            queryset = mappers.get_subscriptions_by_subscriber(instance, queryset)
            queryset.update(subscriber=instance)


def convert_datetime_strings_to_datetimes(data, model):
    datetime_fields = [field.name for field in model._meta.get_fields() if isinstance(field, models.DateTimeField)]
    for field in datetime_fields:
        if field not in data:
            continue

        try:
            data[field] = datetime.strptime(data[field], PADDLE_DATETIME_FORMAT)
        except ValueError:
            # Some fields such as next_bill_date should perhaps be
            # a DateField not DatetimeField
            data[field] = datetime.strptime(data[field], PADDLE_DATE_FORMAT)

        if djsettings.USE_TZ:
            local_time_zone = timezone.get_default_timezone()
            data[field] = timezone.make_aware(data[field], local_time_zone)

    return data



import base64
import collections

import phpserialize
from Crypto.PublicKey import RSA

try:
    from Crypto.Hash import SHA1
except ImportError:  # pragma: no cover
    from Crypto.Hash import SHA as SHA1

try:
    from Crypto.Signature import PKCS1_v1_5
except ImportError:  # pragma: no cover
    from Crypto.Signature import pkcs1_15 as PKCS1_v1_5


PADDLE_DATE_FORMAT = "%Y-%m-%d"
PADDLE_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"


def convert_pubkey_to_rsa(key):
    """
    convert key from PEM to DER - Strip the first and last lines and newlines,
    decode and return an instance of `Crypto.PublicKey.RSA`
    """
    public_key_encoded = "".join(key.split("\n")[1:-1])
    public_key_der = base64.b64decode(public_key_encoded)
    return RSA.importKey(public_key_der)


def is_valid_webhook(payload):
    data = dict(payload)
    signature = data.pop("p_signature", None)
    if signature is None:
        return False

    # Ensure all the data fields are strings
    for field in data:
        data[field] = str(data[field])

    # Sort the data and serialize it
    sorted_data = collections.OrderedDict(sorted(data.items()))
    serialized_data = phpserialize.dumps(sorted_data)

    # verify the data
    digest = SHA1.new()
    digest.update(serialized_data)
    signature = base64.b64decode(signature)

    from . import settings

    verifier = PKCS1_v1_5.new(settings.DJPADDLE_KEY)

    return verifier.verify(digest, signature)



from django.contrib import admin

from . import models

admin.site.register(models.Checkout)


class PriceInline(admin.TabularInline):
    model = models.Price


@admin.register(models.Plan)
class PlanAdmin(admin.ModelAdmin):
    inlines = (PriceInline,)
    list_display = (
        "id",
        "name",
    )


@admin.register(models.Subscription)
class SubscriptionAdmin(admin.ModelAdmin):
    list_display = (
        "subscriber",
        "email",
        "status",
        "plan",
    )
    list_filter = (
        "status",
        "plan",
    )



from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .utils import convert_pubkey_to_rsa

# can be found at https://vendors.paddle.com/authentication
DJPADDLE_VENDOR_ID = getattr(settings, "DJPADDLE_VENDOR_ID")
if not DJPADDLE_VENDOR_ID:
    raise ImproperlyConfigured("'DJPADDLE_VENDOR_ID' must be set")
# create one at https://vendors.paddle.com/authentication
DJPADDLE_API_KEY = getattr(settings, "DJPADDLE_API_KEY")
if not DJPADDLE_API_KEY:
    raise ImproperlyConfigured("'DJPADDLE_API_KEY' must be set")

# can be found at https://vendors.paddle.com/public-key
DJPADDLE_PUBLIC_KEY = getattr(settings, "DJPADDLE_PUBLIC_KEY")
if not DJPADDLE_PUBLIC_KEY:
    raise ImproperlyConfigured("'DJPADDLE_PUBLIC_KEY' must be set")
try:
    DJPADDLE_KEY = convert_pubkey_to_rsa(DJPADDLE_PUBLIC_KEY)
except Exception as e:
    msg = "failed to convert 'DJPADDLE_PUBLIC_KEY'; original message: {}".format(e)
    raise ImproperlyConfigured(msg)

DJPADDLE_SUBSCRIBER_MODEL = getattr(settings, "DJPADDLE_SUBSCRIBER_MODEL", settings.AUTH_USER_MODEL)

DJPADDLE_LINK_STALE_SUBSCRIPTIONS = getattr(settings, "DJPADDLE_LINK_STALE_SUBSCRIPTIONS", True)


DJPADDLE_SUBSCRIBER_BY_PAYLOAD = getattr(
    settings, "DJPADDLE_SUBSCRIBER_BY_PAYLOAD", "djpaddle.mappers.subscriber_by_payload"
)
DJPADDLE_SUBSCRIPTIONS_BY_SUBSCRIBER = getattr(
    settings,
    "DJPADDLE_SUBSCRIPTIONS_BY_SUBSCRIBER",
    "djpaddle.mappers.subscriptions_by_subscriber",
)


def get_subscriber_model():
    """
    Attempt to pull settings.DJPADDLE_SUBSCRIBER_MODEL.
    Users have the option of specifying a custom subscriber model via the
    DJPADDLE_SUBSCRIBER_MODEL setting.
    This methods falls back to AUTH_USER_MODEL if DJPADDLE_SUBSCRIBER_MODEL is not set.
    Returns the subscriber model that is active in this project.
    """
    model_name = DJPADDLE_SUBSCRIBER_MODEL

    # Attempt a Django 1.7 app lookup
    try:
        subscriber_model = django_apps.get_model(model_name)
    except ValueError:
        raise ImproperlyConfigured("DJPADDLE_SUBSCRIBER_MODEL must be of the form 'app_label.model_name'.")
    except LookupError:
        raise ImproperlyConfigured(
            "DJPADDLE_SUBSCRIBER_MODEL refers to model '{model}' "
            "that has not been installed.".format(model=model_name)
        )

    return subscriber_model



import importlib
import threading

from . import settings

_cache = threading.local()
_cache.modules = {}


def _get_fn(fn, *args, **kwargs):
    mod_name, func_name = fn.rsplit(".", 1)
    try:
        cached_modules = _cache.modules
    except AttributeError:  # pragma: no cover
        # AttributeError: '_thread._local' object has no attribute 'modules'
        cached_modules = {}

    if mod_name not in cached_modules:
        cached_modules[mod_name] = importlib.import_module(mod_name)
    return getattr(cached_modules[mod_name], func_name)


def subscriber_by_payload(Subscriber, payload):
    """
    Map a subscriber to a subscription based on the given webhook payload.
    You can overwrite this function via settings.DJPADDLE_SUBSCRIBER_BY_PAYLOAD.

    must throw Subscriber.DoesNotExist in case the mapping failed.
    """
    if "email" not in payload:
        raise Subscriber.DoesNotExist("'email' missing in payload")
    return Subscriber.objects.get(email=payload["email"])


def subscriptions_by_subscriber(subscriber, queryset):
    """
    Filter subscriptions by subscriber. This function is used in order to
    find and link stale subscriptions. You can overwrite this function
    via settings.DJPADDLE_SUBSCRIPTIONS_BY_SUBSCRIBER.
    """
    return queryset.filter(email__iexact=subscriber.email)


def get_subscriber_by_payload(Subscriber, payload):
    """
    wrapper to retrieve and call the function referenced in settings.DJPADDLE_SUBSCRIBER_BY_PAYLOAD
    """
    return _get_fn(settings.DJPADDLE_SUBSCRIBER_BY_PAYLOAD)(Subscriber=Subscriber, payload=payload)


def get_subscriptions_by_subscriber(subscriber, queryset):
    """
    wrapper to retrieve and call the function referenced in settings.DJPADDLE_SUBSCRIPTIONS_BY_SUBSCRIBER
    """
    return _get_fn(settings.DJPADDLE_SUBSCRIPTIONS_BY_SUBSCRIBER)(subscriber=subscriber, queryset=queryset)



# Generated by Django 3.1a1 on 2020-05-22 23:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djpaddle', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Checkout',
            fields=[
                ('id', models.CharField(max_length=40, primary_key=True, serialize=False)),
                ('completed', models.NullBooleanField()),
                ('passthrough', models.TextField(blank=True, null=True)),
                ('email', models.EmailField(blank=True, max_length=254, null=True)),
                ('created_at', models.DateTimeField(blank=True, null=True)),
            ],
        ),
    ]



# Generated by Django 3.1.5 on 2021-01-19 04:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djpaddle', '0003_increase_length_of_paddle_ids'),
    ]

    operations = [
        migrations.AlterField(
            model_name='checkout',
            name='completed',
            field=models.BooleanField(null=True),
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-05 20:22

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import djpaddle.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Plan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('billing_type', models.CharField(choices=[('day', 'day'), ('month', 'month'), ('year', 'year')], max_length=255)),
                ('billing_period', models.IntegerField()),
                ('trial_days', models.IntegerField(default=0)),
            ],
            options={
                'ordering': ['created_at'],
            },
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('id', models.CharField(max_length=32, primary_key=True, serialize=False)),
                ('cancel_url', models.URLField()),
                ('checkout_id', models.CharField(max_length=32)),
                ('currency', models.CharField(max_length=3)),
                ('email', models.EmailField(max_length=254)),
                ('event_time', models.DateTimeField()),
                ('marketing_consent', models.BooleanField()),
                ('next_bill_date', models.DateTimeField()),
                ('passthrough', models.TextField()),
                ('quantity', models.IntegerField()),
                ('source', models.URLField()),
                ('status', models.CharField(choices=[('active', 'active'), ('trialing', 'trialing'), ('past_due', 'past due'), ('paused', 'paused'), ('deleted', 'deleted')], max_length=16)),
                ('unit_price', models.FloatField()),
                ('update_url', models.URLField()),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djpaddle.Plan')),
                ('subscriber', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['created_at'],
            },
        ),
        migrations.CreateModel(
            name='Price',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('currency', djpaddle.fields.PaddleCurrencyCodeField(help_text='Three-letter ISO currency code', max_length=3)),
                ('quantity', models.FloatField()),
                ('recurring', models.BooleanField()),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='prices', to='djpaddle.Plan')),
            ],
            options={
                'ordering': ['currency', 'recurring'],
                'unique_together': {('plan', 'currency', 'recurring')},
            },
        ),
    ]



from django.urls import path

from . import views

app_name = "djpaddle"

urlpatterns = [
    path("webhook/", views.paddle_webhook_view, name="webhook"),
    path("post-checkout/", views.post_checkout_api_view, name="post_checkout_api"),
]
